//TODO: free unused memory from the trees within the print functions
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "gauss.h"
#include "moves.h"
#include "poly.h"

typedef struct intlist { // linked list for storing codes associated with poly
  struct intlist *next;
  int n;
} intlist;

typedef struct polytree { // binary tree for storing polynomials
  struct polytree *left, *right;
  poly p;
  intlist *codes;
} polytree;

typedef struct twopolytree { // binary tree for storing polynomials
  struct twopolytree *left, *right;
  twopoly p;
  intlist *codes;
} twopolytree;

void addtotwotree(twopolytree *, twopoly, int);
int printtwotree(twopolytree *, code *, int);
void addtotree(polytree *, poly, int);
int printtree(polytree *, code *, int, int);

int ns;

int main(int argc, char *argv[]) {
  if (argc < 3) {
    printf("Usage: countpolys2 <max crossings> <number of crossings for output> [max cables] [0 for non-cyclic] [1 for reversible]\n");
    return 0;
  }
  int sz = atoi(argv[1]);
  int maxp = atoi(argv[2]);
  ns = (argc >= 4 ? atoi(argv[3]) : 1);
  //int ns = 4;
  cyclic = (argc >= 5 ? atoi(argv[4]) : 1);
  reversible = (argc == 6 ? atoi(argv[5]) : 0);
  initcrossings((2 * ns * ns - ns) * sz);
  initcrossings(sz);
  initpascal((2 * ns * ns - ns) * sz + 2);
  int n;
  code *codes = genallcodes(maxp, &n);
  //int *lowest = findlowest(codes, n, sz);
  //initlowest(FILENAME, 0);
  initlowest(sz);
  n <<= sz - maxp;
  srand(time(0));

  int i, m = 0;
  twopolytree t;
  t.p.n = -1;
  t.left = t.right = NULL;
  for (i = 0; i < n; i++)
    if (getlowest(i) == -1) {
      fprintf(stderr, "%d of %d\r", i, n);
      code c = getcode(codes, sz, i);
      if (c.n != -1) {
	if (c.n <= maxp) {
	  m++;
	  fflush(stdout);
	  //poly p = multijonespoly(c, ns);
	  twopoly p = genalexander(c);
	  if (t.p.n == -1) { // first one
	    t.p = p;
	    t.codes = calloc(1, sizeof(intlist));
	    t.codes->n = i;
	  } else
	    addtotwotree(&t, p, i);
	  //free(p.coeffs);
	}
	free(c.code);
      }
    }
  printf("                          \r");
  int j = printtwotree(&t, codes, sz);
  printf("%d unique codes, with %d different combinations of polynomials\n",
	 m, j);
  return 0;
}

void addtotwotree(twopolytree *t, twopoly p, int n) {
  int cmp = comparetwopoly(p, t->p);
  if (cmp == 0) {
    intlist *l = t->codes;
    while (l->next != NULL)
      l = l->next;
    l->next = malloc(sizeof(intlist));
    l = l->next;
    l->next = NULL;
    l->n = n;
    freetwopoly(p); // this is only called by main(), and it's appropriate here
    return;
  } else if (cmp == -1) {
    if (t->left != NULL) {
      addtotwotree(t->left, p, n);
      return;
    } else {
      t->left = calloc(1, sizeof(twopolytree));
      t = t->left;
    }
  } else {
    if (t->right != NULL) {
      addtotwotree(t->right, p, n);
      return;
    } else {
      t->right = calloc(1, sizeof(twopolytree));
      t = t->right;
    }
  }
  t->p = p;
  t->codes = malloc(sizeof(intlist));
  t->codes->next = NULL;
  t->codes->n = n;
}

void addtotree(polytree *t, poly p, int n) {
  int cmp = comparepoly(p, t->p);
  if (cmp == 0) {
    intlist *l = t->codes;
    while (l->next != NULL)
      l = l->next;
    l->next = malloc(sizeof(intlist));
    l = l->next;
    l->next = NULL;
    l->n = n;
    free(p.coeffs); // this is only called by main(), and it's appropriate here
    return;
  } else if (cmp == -1) {
    if (t->left != NULL) {
      addtotree(t->left, p, n);
      return;
    } else {
      t->left = calloc(1, sizeof(polytree));
      t = t->left;
    }
  } else {
    if (t->right != NULL) {
      addtotree(t->right, p, n);
      return;
    } else {
      t->right = calloc(1, sizeof(polytree));
      t = t->right;
    }
  }
  t->p = p;
  t->codes = malloc(sizeof(intlist));
  t->codes->next = NULL;
  t->codes->n = n;
}

int printtwotree(twopolytree *t, code *codes, int sz) { // returns number of polys
  if (t == NULL)
    return 0;
  int n = 0;
  n += printtwotree(t->left, codes, sz);
  printtwopoly(t->p);
  printf("\n");
  intlist *l;
  if (t->codes->next == NULL) { // just one code
    code c = getcode(codes, sz, t->codes->n);
    printf("\t\t\t");
    printcode(c);
    free(c.code);
    n++;
  } else {
    polytree pt;
    pt.p.n = -1;
    pt.left = pt.right = NULL;
    for (l = t->codes; l != NULL; l = l->next) {
      code c = getcode(codes, sz, l->n);
      //printf("%d \t", Jinvar(c));
      //printcode(c);
      poly p = jonespoly2(c);
      if (pt.p.n == -1) {
	pt.p = p;
	pt.codes = calloc(1, sizeof(intlist));
	pt.codes->n = l->n;
      } else
	addtotree(&pt, p, l->n);
      free(c.code);
    }
    printf("\tJones polynomials:\n");
    n += printtree(&pt, codes, sz, 1);
  }
  n += printtwotree(t->right, codes, sz);
  return n;
}

void freeintlist(intlist *l) {
  if (l == NULL)
    return;
  freeintlist(l->next);
  free(l);
}

int printtree(polytree *t, code *codes, int sz, int nc) { // returns number of polys
  if (t == NULL)
    return 0;
  int n = 0, i;
  n += printtree(t->left, codes, sz, nc);
  for (i = 0; i < nc; i++)
    printf("\t");
  printpoly(t->p);
  printf("\n");
  intlist *l;
  if (t->codes->next == NULL) { // just one code
    code c = getcode(codes, sz, t->codes->n);
    printf("\t\t\t");
    printcode(c);
    free(c.code);
    n++;
  } else if (nc >= ns) {
    for (l = t->codes; l != NULL; l = l->next) {
      code c = getcode(codes, sz, l->n);
      //printf("%d \t", Jinvar(c));
      printf("\t\t\t");
      printcode(c);
      free(c.code);
    }
    n++;
  } else {
    polytree pt;
    pt.p.n = -1;
    pt.left = pt.right = NULL;
    for (l = t->codes; l != NULL; l = l->next) {
      code c = getcode(codes, sz, l->n);
      poly p = multijonespoly(c, nc + 1);
      if (pt.p.n == -1) {
	pt.p = p;
	pt.codes = calloc(1, sizeof(intlist));
	pt.codes->n = l->n;
      } else
	addtotree(&pt, p, l->n);
      free(c.code);
    }
    for (i = 0; i <= nc; i++)
      printf("\t");
    printf("%d cables:\n", nc + 1);
    n += printtree(&pt, codes, sz, nc + 1);
  }
  n += printtree(t->right, codes, sz, nc);

  free(t->p.coeffs);
  freeintlist(t->codes);
  if (t->left != NULL)
    free(t->left);
  if (t->right != NULL)
    free(t->right);
  return n;
}
